'
' RichOle.bas - RichEdit with OLE support.
'------------------------------------------------------------------------------
' Sample shows how to load/save files with embedded pictures/OLE objects,
' and how to insert OLE objects and picture files directly via OpenFile
' and OLE2's InsertObject dialog. It even does right-click object related
' popup menu for opening an embedded object for editing (DoVerb)..
'
' Public Domain by Borje Hagsten, April 2003.
' Free to use at own responsibility. No guarantees that translations
' of declares, etc. are 100% correct, but at least it seems to work.
'
' Also not sure if all AddRef's and Release stuff are there yet. Own
' tests shows no specific leaks or problems, but one never knows. This
' is first outcast - a more comprehensive version will follow, some day..
'
' LOG:
' Apr 14, 2003: Added %ES_AUTOHSCROLL and %WS_CLIPSIBLINGS styles to
' dialog, plus setting of approximate page width with %EM_SETTARGETDEVICE
' under %WM_INITDIALOG. See comments under %WM_INITDIALOG.
'
' Credits to Nick Melnick, who did the ground-breaking work in a sample
' that showed how to nsert pictures via the clipboard and how to use
' the EM_SETOLECALLBACK and vTables to enable object on load. Much of
' this code is based on his work - THANK YOU, Nick! And of course also
' THANK YOU to Don Dickinson for his pioneer work on RTF streaming.
'
' All "working" code is place in an include file called RICHCOM2.INC,
' and sample code for using the routines here, in RichOle.bas. A bit
' long introduction and some explanations can be found in RICHCOM2.INC.
'
' Note: If trying to understand RICHCOM2.INC makes you feel sick or dizzy,
' skip reading and just concentrate on two calls. Reo_SetComInterface to
' initate RichEdit's COM interface (see WM_INITDIALOG) and Reo_InsertObject
' to insert a picture file or an object via OLE InsertObject dialog. That's
' all you actually need to be able to build a quite advanced Word Processor.. :-)
'
' BTW, when it comes to inserting pictures, all kinds of formats can be
' inserted, as long as they are supported by some good graphics server.
' I've only listed bmp's in Insert dialog, but you can use "All files" (*.*)
' and test to insert other formats too.
'------------------------------------------------------------------------------
#COMPILE EXE
#INCLUDE "WIN32API.INC"
#INCLUDE "COMDLG32.INC"
#INCLUDE "RICHEDIT.INC"
'------------------------------------------------------------------------------
#INCLUDE "RICHOLE2.INC"
'------------------------------------------------------------------------------
%IDC_RICHEDIT  = 150
%IDM_OPEN      = 220 'Note - also se some Edit menu equates in RICHOLE2.INC ..
%IDM_SAVE      = 221
%IDM_INSIMG    = 230
%IDM_INSOBJ    = 231
'------------------------------------------------------------------------------
DECLARE CALLBACK FUNCTION DlgProc() AS LONG
DECLARE FUNCTION RichEdit_LoadFile (BYVAL hEdit AS DWORD, BYVAL sFileName AS STRING) AS LONG
DECLARE FUNCTION RichEdit_SaveFile (BYVAL hEdit AS DWORD, BYVAL sFileName AS STRING) AS LONG
'------------------------------------------------------------------------------

'
' Main entrance - create dialog and controls, etc
'------------------------------------------------------------------------------
FUNCTION PBMAIN () AS LONG
  LOCAL hDlg AS LONG 'should be DWORD, but to enable compile with PB/DLL 6.x ..
 
  DIALOG NEW 0, "RichOle sample - Document 1",,, 380, 240, %WS_OVERLAPPEDWINDOW, 0 TO hDlg
 
  CONTROL ADD BUTTON, hDlg, %IDM_OPEN,   "&Open file",      2, 2, 50, 14
  CONTROL ADD BUTTON, hDlg, %IDM_SAVE,   "&Save file",     52, 2, 50, 14
  CONTROL ADD BUTTON, hDlg, %IDM_INSIMG, "Insert Image",  106, 2, 50, 14
  CONTROL ADD BUTTON, hDlg, %IDM_INSOBJ, "Insert Object", 156, 2, 50, 14
  CONTROL ADD BUTTON, hDlg, %IDCANCEL,   "E&xit",         210, 2, 50, 14
 
  IF LoadLibrary("RICHED32.DLL") THEN
     CONTROL ADD "RichEdit", hDlg, %IDC_RICHEDIT, "", 0, 18, 380, 190, _
                             %WS_CHILD OR %WS_CLIPSIBLINGS OR %WS_VISIBLE OR _
                             %WS_HSCROLL OR %WS_VSCROLL OR %ES_MULTILINE OR _
                             %ES_AUTOHSCROLL OR %ES_AUTOVSCROLL OR %ES_NOHIDESEL OR _
                             %ES_SAVESEL OR %ES_WANTRETURN, %WS_EX_CLIENTEDGE
  ELSE
     EXIT FUNCTION
  END IF
 
  CONTROL SEND hDlg, %IDC_RICHEDIT, %EM_SETOPTIONS, %ECOOP_OR, %ECO_SELECTIONBAR 'for left margin select..
  CONTROL SEND hDlg, %IDC_RICHEDIT, %EM_EXLIMITTEXT,  0, 1024 * 1024 - 1 '1 MB limit in Win95/98?
 
  DIALOG SHOW MODAL hDlg CALL DlgProc
END FUNCTION
 
'
' Main dialog's callback procedure
'------------------------------------------------------------------------------
CALLBACK FUNCTION DlgProc() AS LONG
  SELECT CASE CBMSG
     CASE %WM_INITDIALOG 'first message - initiate
        LOCAL dwRes AS DWORD, fName AS STRING, sBuf AS STRING, rc AS RECT
        STATIC hEdit AS DWORD, sPath AS STRING, sFile AS STRING

        hEdit = GetDlgItem(CBHNDL, %IDC_RICHEDIT)
        sPath = CURDIR$
        IF RIGHT$(sPath, 1) <> "\" THEN sPath = sPath + "\"
        sFile = "Document 1"
        Reo_SetComInterface hEdit 'initiate RichEdit's COM interface

        ' Set page width for proper WYSIWYG
        ' Should grab data from printer DC, but here's a quick cheat.
        ' Note: measurement is twips, where 1440 twips is an inch,
        ' and 567 twips is a centimeter.
        ' Letter/Legal width: 8.5 inch x 1440 = 12240 twips
        ' A4 width:            21 cm   x  567 = 11907 twips
        ' Minus margins, for example an inch each for left/right margin.
        '
        ' But - there are additional problems in RichEdit. For some reason,
        ' following MS advice do not work. Via compare, I have found that one
        ' has to decrease result with about 0.8 cm for proper WYSIWYG. Reason
        ' is probably related to non-printable areas, which RichEdit for some
        ' weird reason seems to add by itself. When grabbing measures from
        ' printer, it's easier to get exact result, but here we simply guess
        ' that decreasing result with (0.8 * 567) will do it:

        dwRes = GetDC(hEdit)
'           SendMessage hEdit, %EM_SETTARGETDEVICE, dwRes, 11907 - (4 * 567) - (0.8 * 567) 'Sweden, etc..
           SendMessage hEdit, %EM_SETTARGETDEVICE, dwRes, 12240 - (2 * 1440) - (0.8 * 567) 'USA
        ReleaseDC hEdit, dwRes

     CASE %WM_COMMAND
        SELECT CASE CBCTL
           CASE %IDM_OPEN
              IF CBCTLMSG = %BN_CLICKED THEN
                 fName = "*.RTF"
                 sBuf  = "Rich Text files (*.RTF)|*.rtf|" + _
                         "Text files (*.TXT)|*.txt|" + _
                         "All Files (*.*)|*.*"
                 dwRes = %OFN_EXPLORER OR %OFN_FILEMUSTEXIST OR %OFN_HIDEREADONLY
                 IF OpenFileDialog(CBHNDL, "", fName, sPath, sBuf, "RTF", dwRes) THEN
                    sFile = MID$(fName, INSTR(-1, fName, ANY "\/") + 1)
                    sPath = LEFT$(fName, INSTR(-1, fName, ANY "\/"))

                    RichEdit_LoadFile hEdit, fName 'load file into RichEdit

                    sBuf = "RichOle - " + fName
                    SetWindowText CBHNDL, BYVAL STRPTR(sBuf)
                 END IF
              END IF

           CASE %IDM_SAVE
              IF CBCTLMSG = %BN_CLICKED THEN
                 fName = sFile
                 dwRes = %OFN_EXPLORER OR %OFN_FILEMUSTEXIST OR _
                         %OFN_HIDEREADONLY OR %OFN_OVERWRITEPROMPT
                 sBuf  = "Rich Text   (*.RTF)|*.rtf|" + _
                         "Text Files   (*.TXT)|*.txt|" + _
                         "All Files|*.*"
                 IF SaveFileDialog(CBHNDL, "", fName, sPath, sBuf, "RTF", dwRes) THEN
                    sFile = MID$(fName, INSTR(-1, fName, ANY "\/") + 1) 'in case it was changed..
                    sPath = LEFT$(fName, INSTR(-1, fName, ANY "\/"))

                    RichEdit_SaveFile hEdit, fName 'save file from RichEdit

                    sBuf = "RichOle - " + fName 'show file name in dialog title
                    SetWindowText CBHNDL, BYVAL STRPTR(sBuf)
                 END IF
              END IF

           CASE %IDM_INSIMG
              IF CBCTLMSG = %BN_CLICKED THEN
                 fName   = "*.bmp"
                 sBuf    = "Bitmap picture (*.BMP)|*.bmp|" + "All Files (*.*)|*.*"
                 dwRes = %OFN_FILEMUSTEXIST OR %OFN_HIDEREADONLY OR %OFN_LONGNAMES
                 IF OpenFileDialog(CBHNDL, "", fName, sPath, sBuf, "bmp", dwRes) THEN
                    Reo_InsertObject hEdit, fName, "RichCom", sFile
                 END IF
              END IF

           CASE %IDM_INSOBJ
              IF CBCTLMSG = %BN_CLICKED THEN
                 MOUSEPTR 11
                 ' Note: Second parameter is file name. If no file name is given,
                 ' the function shows OLE InsertObject dialog instead. 3'd and 4'th
                 ' parameters are Program and Doc -titles that later will be shown
                 ' in an object's server at editing (Close and return to..). Last
                 ' one, Doc title, seems to be the most important one..
                 Reo_InsertObject hEdit, "", "RichCom", "RichCom Document"
              END IF

           CASE %IDCANCEL 'Exit
              IF CBCTLMSG = %BN_CLICKED THEN DIALOG END CBHNDL

          'EDIT MENU COMMANDS
           CASE %IDM_UNDO 'Undo last action
              IF CBCTLMSG = %BN_CLICKED THEN
                 SendMessage hEdit, %EM_UNDO, 0, 0 
              END IF

           CASE %IDM_CUT 'cut out and copy selection
              IF CBCTLMSG = %BN_CLICKED THEN
                 SendMessage hEdit, %WM_CUT, 0, 0
              END IF

           CASE %IDM_COPY 'copy selection
              IF CBCTLMSG = %BN_CLICKED THEN
                 SendMessage hEdit, %WM_COPY, 0, 0
              END IF

           CASE %IDM_PASTE 'Paste clipboard contents into RichEdit
              IF CBCTLMSG = %BN_CLICKED THEN
                 SendMessage hEdit, %WM_PASTE, 0, 0
              END IF

           CASE %IDM_DELETE 'delete selection
              IF CBCTLMSG = %BN_CLICKED THEN
                 SendMessage hEdit, %WM_CLEAR, 0, 0
              END IF

           CASE %IDM_SELALL 'Select all
              IF CBCTLMSG = %BN_CLICKED THEN
                 SendMessage hEdit, %EM_SETSEL, 0, -1
              END IF
           '------------------------------------------------

           CASE >= %IDM_OBJECTMIN 'this is for custom verbs, for objects..
              IF CBCTLMSG = %BN_CLICKED THEN
                 Reo_DoVerb hEdit, CBCTL - %IDM_OBJECTMIN
              END IF

        END SELECT
 
     CASE %WM_SIZE 'resize message
        IF CBWPARAM <> %SIZE_MINIMIZED THEN
           GetWindowRect hEdit, rc           ' get Richedit's position on screen
           MapWindowPoints 0, CBHNDL, rc, 2  ' map rect to dialog - we need top pos
           SetWindowPos hEdit, 0, 0, 0, _
                        LOWRD(CBLPARAM), HIWRD(CBLPARAM) - rc.nTop, _
                        %SWP_NOMOVE OR %SWP_NOZORDER
        END IF
 
  END SELECT
END FUNCTION
 
'
' Load a file into a RichEdit control
'------------------------------------------------------------------------------
FUNCTION RichEdit_LoadFile (BYVAL hEdit AS DWORD, BYVAL sFileName AS STRING) AS LONG

  LOCAL ff AS DWORD, txt AS STRING, eStream AS EDITSTREAM

  ff = FREEFILE
  OPEN sFileName FOR BINARY AS ff LEN = 8192

  IF ERR = 0 THEN
     IF UCASE$(RIGHT$(sFileName, 4)) = ".RTF" THEN
        eStream.dwCookie = ff
        eStream.pfnCallback = CODEPTR(RichEditStreamInFile)
        eStream.dwError = 0

        SendMessage hEdit, %EM_STREAMIN, %SF_RTF, VARPTR(eStream)
     ELSE
        GET$ ff, LOF(ff), txt  'probably text file, read entire contents into text buffer
        SendMessage hEdit, %WM_SETTEXT, 0, STRPTR(txt) 'set text and exit
     END IF

     CLOSE ff
     FUNCTION = %TRUE

  ELSE 'on file I/O error
     MessageBeep &HFFFFFFFF  'I prefer speaker beep on errors..
     RESET : ERRCLEAR        'Should of course flash a message here, but is sample code..
  END IF

END FUNCTION

'
' Rich Edit stream in callback - for streaming in file contents
'------------------------------------------------------------------------------
FUNCTION RichEditStreamInFile (BYVAL dwCookie AS DWORD, pbBuff AS BYTE PTR, _
                               BYVAL cb AS LONG, pcb AS LONG) AS DWORD

  LOCAL sInput AS STRING
  
  'Find out if the amount requested is beyond the end of the file.
  IF LOF(dwCookie) - LOC(dwCookie) < 0 THEN
     cb = LOF(dwCookie) - LOC(dwCookie)
  END IF
  sInput = SPACE$(cb)

  'Incoming data
  GET #dwCookie,, sInput
  IF ERR THEN EXIT FUNCTION

  pcb = cb
  POKE$ pbBuff, sInput

END FUNCTION

'
' Save a file from a RichEdit control
'------------------------------------------------------------------------------
FUNCTION RichEdit_SaveFile(BYVAL hEdit AS DWORD, BYVAL sFileName AS STRING ) AS LONG
  LOCAL ff AS DWORD, lRes AS LONG, txt AS STRING, eStream AS EDITSTREAM

  ON ERROR RESUME NEXT
  ff = FREEFILE
  OPEN sFileName FOR BINARY AS ff LEN = 8192

  IF ERR = 0 THEN
     IF UCASE$(RIGHT$(sFileName, 4)) = ".RTF" THEN
        eStream.dwCookie = ff
        eStream.dwError = 0
        eStream.pfnCallback = CODEPTR(RichEditStreamToFile)
        SendMessage hEdit, %EM_STREAMOUT, %SF_RTF, VARPTR(eStream)

     ELSE 'probably text file, simply grab text and save directly
        lRes = SendMessage(hEdit, %WM_GETTEXTLENGTH, 0, 0)
        txt = SPACE$(lRes + 1)
        SendMessage hEdit, %WM_GETTEXT, lRes, STRPTR(txt)
        txt = LEFT$(txt, LEN(txt) - 1)
        PUT$ ff, txt  'write txt to disk
     END IF

  ELSE 'on file I/O error
     MessageBeep &HFFFFFFFF  'I prefer speaker beep on errors..
     RESET : ERRCLEAR        'Should of course flash a message here, but is sample code..
     EXIT FUNCTION
  END IF

  SETEOF ff
  CLOSE ff
  FUNCTION = %TRUE 'return %TRUE on success

END FUNCTION
 
'
' Rich Edit stream out callback - for streaming out to file
'------------------------------------------------------------------------------
FUNCTION RichEditStreamToFile (BYVAL dwCookie AS DWORD, pbBuffer AS BYTE PTR, _
                               BYVAL cb AS LONG, pcb AS LONG) AS DWORD
  LOCAL tmpStr AS STRING
  tmpStr = PEEK$(pbBuffer, cb)
  PUT #dwCookie,, tmpStr
END FUNCTION
